@@ -88,29 +88,45 @@ assert_eq!(group.as_str(), "20");
 
 # Syntax
 
-The regex syntax is based on the [regex] crate's, with some additional supported syntax. Escapes:
+The regex syntax is based on the [regex] crate's, with some additional supported syntax.
 
-```norun
-\h    hex digit ([0-9A-Fa-f])
-\H    not hex digit ([^0-9A-Fa-f])
-\e    escape control character (\x1B)
-```
+Escapes:
+
+`\h`
+: hex digit (`[0-9A-Fa-f]`) \
+`\H`
+: not hex digit (`[^0-9A-Fa-f]`) \
+`\e`
+: escape control character (`\x1B`)
 
 Backreferences:
 
-```norun
-\1    match the exact string that the first capture group matched
-\2    backref to the second capture group, etc
-```
+`\1`
+: match the exact string that the first capture group matched \
+`\2`
+: backref to the second capture group, etc
+
+Named capture groups:
+
+`(?<name>exp)`
+: match *exp*, creating capture group named *name* \
+`\k<name>`
+: match the exact string that the capture group named *name* matched \
+`(?P<name>exp)`
+: same as `(?<name>exp)` for compatibility with Python, etc. \
+`(?P=name)`
+: same as `\k<name>` for compatibility with Python, etc.
 
 Look-around assertions for matching without changing the current position:
 
-```norun
-(?=exp)    look-ahead, succeeds if exp matches to the right of the current position
-(?!exp)    negative look-ahead, succeeds if exp doesn't match to the right
-(?<=exp)   look-behind, succeeds if exp matches to the left of the current position
-(?<!exp)   negative look-behind, succeeds if exp doesn't match to the left
-```
+`(?=exp)`
+: look-ahead, succeeds if *exp* matches to the right of the current position \
+`(?!exp)`
+: negative look-ahead, succeeds if *exp* doesn't match to the right \
+`(?<=exp)`
+: look-behind, succeeds if *exp* matches to the left of the current position \
+`(?<!exp)`
+: negative look-behind, succeeds if *exp* doesn't match to the left
 
 Atomic groups using `(?>exp)` to prevent backtracking within `exp`, e.g.:
 
@@ -125,26 +141,32 @@ assert!(!re.is_match("abc").unwrap());
 [regex]: https://crates.io/crates/regex
 */
 
-#![doc(html_root_url = "https://docs.rs/fancy-regex/0.3.5")]
+#![doc(html_root_url = "https://docs.rs/fancy-regex/0.6.0")]
 #![deny(missing_docs)]
 #![deny(missing_debug_implementations)]
 
-use bit_set::BitSet;
 use std::fmt;
+use std::fmt::{Debug, Formatter};
+use std::ops::{Index, Range};
+use std::str::FromStr;
+use std::sync::Arc;
 use std::usize;
 
 mod analyze;
 mod compile;
 mod error;
+mod expand;
 mod parse;
 mod vm;
 
 use crate::analyze::analyze;
 use crate::compile::compile;
-use crate::parse::Parser;
+use crate::parse::{ExprTree, NamedGroups, Parser};
 use crate::vm::Prog;
 
 pub use crate::error::{Error, Result};
+pub use crate::expand::Expander;
+use std::borrow::Cow;
 
 const MAX_RECURSION: usize = 64;
 
@@ -155,9 +177,14 @@ const MAX_RECURSION: usize = 64;
 pub struct RegexBuilder(RegexOptions);
 
 /// A compiled regular expression.
-pub struct Regex(RegexImpl);
+#[derive(Clone)]
+pub struct Regex {
+    inner: RegexImpl,
+    named_groups: Arc<NamedGroups>,
+}
 
 // Separate enum because we don't want to expose any of this
+#[derive(Clone)]
 enum RegexImpl {
     // Do we want to box this? It's pretty big...
     Wrap {
@@ -179,9 +206,131 @@ pub struct Match<'t> {
     end: usize,
 }
 
+/// An iterator over all non-overlapping matches for a particular string.
+///
+/// The iterator yields a `Result<Match>`. The iterator stops when no more
+/// matches can be found.
+///
+/// `'r` is the lifetime of the compiled regular expression and `'t` is the
+/// lifetime of the matched string.
+#[derive(Debug)]
+pub struct Matches<'r, 't> {
+    re: &'r Regex,
+    text: &'t str,
+    last_end: usize,
+    last_match: Option<usize>,
+}
+
+impl<'r, 't> Matches<'r, 't> {
+    /// Return the text being searched.
+    pub fn text(&self) -> &'t str {
+        self.text
+    }
+
+    /// Return the underlying regex.
+    pub fn regex(&self) -> &'r Regex {
+        &self.re
+    }
+}
+
+impl<'r, 't> Iterator for Matches<'r, 't> {
+    type Item = Result<Match<'t>>;
+
+    /// Adapted from the `regex` crate. Calls `find_from_pos` repeatedly.
+    /// Ignores empty matches immediately after a match.
+    fn next(&mut self) -> Option<Self::Item> {
+        if self.last_end > self.text.len() {
+            return None;
+        }
+
+        let mat = match self.re.find_from_pos(self.text, self.last_end) {
+            Err(error) => return Some(Err(error)),
+            Ok(None) => return None,
+            Ok(Some(mat)) => mat,
+        };
+
+        if mat.start == mat.end {
+            // This is an empty match. To ensure we make progress, start
+            // the next search at the smallest possible starting position
+            // of the next match following this one.
+            self.last_end = next_utf8(self.text, mat.end);
+            // Don't accept empty matches immediately following a match.
+            // Just move on to the next match.
+            if Some(mat.end) == self.last_match {
+                return self.next();
+            }
+        } else {
+            self.last_end = mat.end;
+        }
+
+        self.last_match = Some(mat.end);
+
+        Some(Ok(mat))
+    }
+}
+
+/// An iterator that yields all non-overlapping capture groups matching a
+/// particular regular expression.
+///
+/// The iterator stops when no more matches can be found.
+///
+/// `'r` is the lifetime of the compiled regular expression and `'t` is the
+/// lifetime of the matched string.
+#[derive(Debug)]
+pub struct CaptureMatches<'r, 't>(Matches<'r, 't>);
+
+impl<'r, 't> CaptureMatches<'r, 't> {
+    /// Return the text being searched.
+    pub fn text(&self) -> &'t str {
+        self.0.text
+    }
+
+    /// Return the underlying regex.
+    pub fn regex(&self) -> &'r Regex {
+        &self.0.re
+    }
+}
+
+impl<'r, 't> Iterator for CaptureMatches<'r, 't> {
+    type Item = Result<Captures<'t>>;
+
+    /// Adapted from the `regex` crate. Calls `captures_from_pos` repeatedly.
+    /// Ignores empty matches immediately after a match.
+    fn next(&mut self) -> Option<Self::Item> {
+        if self.0.last_end > self.0.text.len() {
+            return None;
+        }
+
+        let captures = match self.0.re.captures_from_pos(self.0.text, self.0.last_end) {
+            Err(error) => return Some(Err(error)),
+            Ok(None) => return None,
+            Ok(Some(captures)) => captures,
+        };
+
+        let mat = captures
+            .get(0)
+            .expect("`Captures` is expected to have entire match at 0th position");
+        if mat.start == mat.end {
+            self.0.last_end = next_utf8(self.0.text, mat.end);
+            if Some(mat.end) == self.0.last_match {
+                return self.next();
+            }
+        } else {
+            self.0.last_end = mat.end;
+        }
+
+        self.0.last_match = Some(mat.end);
+
+        Some(Ok(captures))
+    }
+}
+
 /// A set of capture groups found for a regex.
 #[derive(Debug)]
-pub struct Captures<'t>(CapturesImpl<'t>);
+pub struct Captures<'t> {
+    inner: CapturesImpl<'t>,
+    named_groups: Arc<NamedGroups>,
+}
 
 #[derive(Debug)]
 enum CapturesImpl<'t> {
@@ -278,6 +427,22 @@ impl fmt::Debug for Regex {
     }
 }
 
+impl fmt::Display for Regex {
+    /// Shows the original regular expression
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        write!(f, "{}", self.as_str())
+    }
+}
+
+impl FromStr for Regex {
+    type Err = Error;
+
+    /// Attempts to parse a string into a regular expression
+    fn from_str(s: &str) -> Result<Regex> {
+        Regex::new(s)
+    }
+}
+
 impl Regex {
     /// Parse and compile a regex with default options, see `RegexBuilder`.
     ///
@@ -291,21 +456,24 @@ impl Regex {
     }
 
     fn new_options(options: RegexOptions) -> Result<Regex> {
-        let (raw_e, backrefs) = Expr::parse(&options.pattern)?;
+        let raw_tree = Expr::parse_tree(&options.pattern)?;
 
         // wrapper to search for re at arbitrary start position,
         // and to capture the match bounds
-        let e = Expr::Concat(vec![
-            Expr::Repeat {
-                child: Box::new(Expr::Any { newline: true }),
-                lo: 0,
-                hi: usize::MAX,
-                greedy: false,
-            },
-            Expr::Group(Box::new(raw_e)),
-        ]);
+        let tree = ExprTree {
+            expr: Expr::Concat(vec![
+                Expr::Repeat {
+                    child: Box::new(Expr::Any { newline: true }),
+                    lo: 0,
+                    hi: usize::MAX,
+                    greedy: false,
+                },
+                Expr::Group(Box::new(raw_tree.expr)),
+            ]),
+            ..raw_tree
+        };
 
-        let info = analyze(&e, &backrefs)?;
+        let info = analyze(&tree)?;
 
         let inner_info = &info.children[1].children[0]; // references inner expr
         if !inner_info.hard {
@@ -313,8 +481,8 @@ impl Regex {
 
             // we do our own to_str because escapes are different
             let mut re_cooked = String::new();
-            // same as raw_e above, but it was moved, so traverse to find it
-            let raw_e = match e {
+            // same as raw_tree.expr above, but it was moved, so traverse to find it
+            let raw_e = match tree.expr {
                 Expr::Concat(ref v) => match v[1] {
                     Expr::Group(ref child) => child,
                     _ => unreachable!(),
@@ -323,20 +491,26 @@ impl Regex {
             };
             raw_e.to_str(&mut re_cooked, 0);
             let inner = compile::compile_inner(&re_cooked, &options)?;
-            return Ok(Regex(RegexImpl::Wrap { inner, options }));
+            return Ok(Regex {
+                inner: RegexImpl::Wrap { inner, options },
+                named_groups: Arc::new(tree.named_groups),
+            });
         }
 
         let prog = compile(&info)?;
-        Ok(Regex(RegexImpl::Fancy {
-            prog,
-            n_groups: info.end_group,
-            options,
-        }))
+        Ok(Regex {
+            inner: RegexImpl::Fancy {
+                prog,
+                n_groups: info.end_group,
+                options,
+            },
+            named_groups: Arc::new(tree.named_groups),
+        })
     }
 
     /// Returns the original string of this regex.
     pub fn as_str(&self) -> &str {
-        match &self.0 {
+        match &self.inner {
             RegexImpl::Wrap { options, .. } => &options.pattern,
             RegexImpl::Fancy { options, .. } => &options.pattern,
         }
@@ -355,7 +529,7 @@ impl Regex {
     /// assert!(re.is_match("mirror mirror on the wall").unwrap());
     /// ```
     pub fn is_match(&self, text: &str) -> Result<bool> {
-        match &self.0 {
+        match &self.inner {
             RegexImpl::Wrap { ref inner, .. } => Ok(inner.is_match(text)),
             RegexImpl::Fancy {
                 ref prog, options, ..
@@ -366,9 +540,37 @@ impl Regex {
         }
     }
 
+    /// Returns an iterator for each successive non-overlapping match in `text`.
+    ///
+    /// If you have capturing groups in your regex that you want to extract, use the [Regex::captures_iter()]
+    /// method.
+    ///
+    /// # Example
+    ///
+    /// Find all words followed by an exclamation point:
+    ///
+    /// ```rust
+    /// # use fancy_regex::Regex;
+    ///
+    /// let re = Regex::new(r"\w+(?=!)").unwrap();
+    /// let mut matches = re.find_iter("so fancy! even with! iterators!");
+    /// assert_eq!(matches.next().unwrap().unwrap().as_str(), "fancy");
+    /// assert_eq!(matches.next().unwrap().unwrap().as_str(), "with");
+    /// assert_eq!(matches.next().unwrap().unwrap().as_str(), "iterators");
+    /// assert!(matches.next().is_none());
+    /// ```
+    pub fn find_iter<'r, 't>(&'r self, text: &'t str) -> Matches<'r, 't> {
+        Matches {
+            re: &self,
+            text,
+            last_end: 0,
+            last_match: None,
+        }
+    }
+
     /// Find the first match in the input text.
     ///
-    /// If you have capturing groups in your regex that you want to extract, use the [captures()]
+    /// If you have capturing groups in your regex that you want to extract, use the [Regex::captures()]
     /// method.
     ///
     /// # Example
@@ -382,17 +584,68 @@ impl Regex {
     /// assert_eq!(re.find("so fancy!").unwrap().unwrap().as_str(), "fancy");
     /// ```
     pub fn find<'t>(&self, text: &'t str) -> Result<Option<Match<'t>>> {
-        match &self.0 {
+        self.find_from_pos(text, 0)
+    }
+
+    /// Returns the first match in `text`, starting from the specified byte position `pos`.
+    ///
+    /// # Examples
+    ///
+    /// Finding match starting at a position:
+    ///
+    /// ```
+    /// # use fancy_regex::Regex;
+    /// let re = Regex::new(r"(?m:^)(\d+)").unwrap();
+    /// let text = "1 test 123\n2 foo";
+    /// let mat = re.find_from_pos(text, 7).unwrap().unwrap();
+    ///
+    /// assert_eq!(mat.start(), 11);
+    /// assert_eq!(mat.end(), 12);
+    /// ```
+    ///
+    /// Note that in some cases this is not the same as using the `find`
+    /// method and passing a slice of the string, see [Regex::captures_from_pos()] for details.
+    pub fn find_from_pos<'t>(&self, text: &'t str, pos: usize) -> Result<Option<Match<'t>>> {
+        match &self.inner {
             RegexImpl::Wrap { inner, .. } => Ok(inner
-                .find(text)
+                .find_at(text, pos)
                 .map(|m| Match::new(text, m.start(), m.end()))),
             RegexImpl::Fancy { prog, options, .. } => {
-                let result = vm::run(prog, text, 0, 0, options)?;
+                let result = vm::run(prog, text, pos, 0, options)?;
                 Ok(result.map(|saves| Match::new(text, saves[0], saves[1])))
             }
         }
     }
 
+    /// Returns an iterator over all the non-overlapping capture groups matched in `text`.
+    ///
+    /// # Examples
+    ///
+    /// Finding all matches and capturing parts of each:
+    ///
+    /// ```rust
+    /// # use fancy_regex::Regex;
+    ///
+    /// let re = Regex::new(r"(\d{4})-(\d{2})").unwrap();
+    /// let text = "It was between 2018-04 and 2020-01";
+    /// let mut all_captures = re.captures_iter(text);
+    ///
+    /// let first = all_captures.next().unwrap().unwrap();
+    /// assert_eq!(first.get(1).unwrap().as_str(), "2018");
+    /// assert_eq!(first.get(2).unwrap().as_str(), "04");
+    /// assert_eq!(first.get(0).unwrap().as_str(), "2018-04");
+    ///
+    /// let second = all_captures.next().unwrap().unwrap();
+    /// assert_eq!(second.get(1).unwrap().as_str(), "2020");
+    /// assert_eq!(second.get(2).unwrap().as_str(), "01");
+    /// assert_eq!(second.get(0).unwrap().as_str(), "2020-01");
+    ///
+    /// assert!(all_captures.next().is_none());
+    /// ```
+    pub fn captures_iter<'r, 't>(&'r self, text: &'t str) -> CaptureMatches<'r, 't> {
+        CaptureMatches(self.find_iter(text))
+    }
+
     /// Returns the capture groups for the first match in `text`.
     ///
     /// If no match is found, then `Ok(None)` is returned.
@@ -437,7 +690,7 @@ impl Regex {
     /// ```
     ///
     /// Note that in some cases this is not the same as using the `captures`
-    /// methods and passing a slice of the string, see the capture that we get
+    /// method and passing a slice of the string, see the capture that we get
     /// when we do this:
     ///
     /// ```
@@ -452,11 +705,15 @@ impl Regex {
     /// of the string slice.
     ///
     pub fn captures_from_pos<'t>(&self, text: &'t str, pos: usize) -> Result<Option<Captures<'t>>> {
-        match &self.0 {
+        let named_groups = self.named_groups.clone();
+        match &self.inner {
             RegexImpl::Wrap { inner, .. } => {
                 let mut locations = inner.capture_locations();
                 let result = inner.captures_read_at(&mut locations, text, pos);
-                Ok(result.map(|_| Captures(CapturesImpl::Wrap { text, locations })))
+                Ok(result.map(|_| Captures {
+                    inner: CapturesImpl::Wrap { text, locations },
+                    named_groups,
+                }))
             }
             RegexImpl::Fancy {
                 prog,
@@ -467,16 +724,37 @@ impl Regex {
                 let result = vm::run(prog, text, pos, 0, options)?;
                 Ok(result.map(|mut saves| {
                     saves.truncate(n_groups * 2);
-                    Captures(CapturesImpl::Fancy { text, saves })
+                    Captures {
+                        inner: CapturesImpl::Fancy { text, saves },
+                        named_groups,
+                    }
                 }))
             }
         }
     }
 
+    /// Returns the number of captures, including the implicit capture of the entire expression.
+    pub fn captures_len(&self) -> usize {
+        match &self.inner {
+            RegexImpl::Wrap { inner, .. } => inner.captures_len(),
+            RegexImpl::Fancy { n_groups, .. } => *n_groups,
+        }
+    }
+
+    /// Returns an iterator over the capture names.
+    pub fn capture_names(&self) -> CaptureNames {
+        let mut names = Vec::new();
+        names.resize(self.captures_len(), None);
+        for (name, &i) in self.named_groups.iter() {
+            names[i] = Some(name.as_str());
+        }
+        CaptureNames(names.into_iter())
+    }
+
     // for debugging only
     #[doc(hidden)]
     pub fn debug_print(&self) {
-        match &self.0 {
+        match &self.inner {
             RegexImpl::Wrap { inner, .. } => println!("wrapped {:?}", inner),
             RegexImpl::Fancy { prog, .. } => prog.debug_print(),
         }
@@ -496,24 +774,44 @@ impl<'t> Match<'t> {
         self.end
     }
 
+    /// Returns the range over the starting and ending byte offsets of the match in text.
+    #[inline]
+    pub fn range(&self) -> Range<usize> {
+        self.start..self.end
+    }
+
     /// Returns the matched text.
     #[inline]
     pub fn as_str(&self) -> &'t str {
         &self.text[self.start..self.end]
     }
 
+    /// Creates a new match from the given text and byte offsets.
     fn new(text: &'t str, start: usize, end: usize) -> Match<'t> {
         Match { text, start, end }
     }
 }
 
+impl<'t> From<Match<'t>> for &'t str {
+    fn from(m: Match<'t>) -> &'t str {
+        m.as_str()
+    }
+}
+
+impl<'t> From<Match<'t>> for Range<usize> {
+    fn from(m: Match<'t>) -> Range<usize> {
+        m.range()
+    }
+}
+
+#[allow(clippy::len_without_is_empty)] // follow regex's API
 impl<'t> Captures<'t> {
     /// Get the capture group by its index in the regex.
     ///
     /// If there is no match for that group or the index does not correspond to a group, `None` is
     /// returned. The index 0 returns the whole match.
     pub fn get(&self, i: usize) -> Option<Match<'t>> {
-        match &self.0 {
+        match &self.inner {
             CapturesImpl::Wrap { text, locations } => {
                 locations
                     .get(i)
@@ -538,21 +836,99 @@ impl<'t> Captures<'t> {
         }
     }
 
+    /// Returns the match for a named capture group.  Returns `None` the capture
+    /// group did not match or if there is no group with the given name.
+    pub fn name(&self, name: &str) -> Option<Match<'t>> {
+        self.named_groups.get(name).and_then(|i| self.get(*i))
+    }
+
+    /// Expands all instances of `$group` in `replacement` to the corresponding
+    /// capture group `name`, and writes them to the `dst` buffer given.
+    ///
+    /// `group` may be an integer corresponding to the index of the
+    /// capture group (counted by order of opening parenthesis where `\0` is the
+    /// entire match) or it can be a name (consisting of letters, digits or
+    /// underscores) corresponding to a named capture group.
+    ///
+    /// If `group` isn't a valid capture group (whether the name doesn't exist
+    /// or isn't a valid index), then it is replaced with the empty string.
+    ///
+    /// The longest possible name is used. e.g., `$1a` looks up the capture
+    /// group named `1a` and not the capture group at index `1`. To exert more
+    /// precise control over the name, use braces, e.g., `${1}a`.
+    ///
+    /// To write a literal `$`, use `$$`.
+    ///
+    /// For more control over expansion, see [`Expander`].
+    ///
+    /// [`Expander`]: expand/struct.Expander.html
+    pub fn expand(&self, replacement: &str, dst: &mut String) {
+        Expander::default().append_expansion(dst, replacement, self);
+    }
+
     /// Iterate over the captured groups in order in which they appeared in the regex. The first
     /// capture corresponds to the whole match.
     pub fn iter<'c>(&'c self) -> SubCaptureMatches<'c, 't> {
         SubCaptureMatches { caps: self, i: 0 }
     }
 
-    /// How many groups were captured.
+    /// How many groups were captured. This is always at least 1 because group 0 returns the whole
+    /// match.
     pub fn len(&self) -> usize {
-        match &self.0 {
+        match &self.inner {
             CapturesImpl::Wrap { locations, .. } => locations.len(),
             CapturesImpl::Fancy { saves, .. } => saves.len() / 2,
         }
     }
 }
 
+/// Copied from [`regex::Captures`]...
+///
+/// Get a group by index.
+///
+/// `'t` is the lifetime of the matched text.
+///
+/// The text can't outlive the `Captures` object if this method is
+/// used, because of how `Index` is defined (normally `a[i]` is part
+/// of `a` and can't outlive it); to do that, use `get()` instead.
+///
+/// # Panics
+///
+/// If there is no group at the given index.
+impl<'t> Index<usize> for Captures<'t> {
+    type Output = str;
+
+    fn index(&self, i: usize) -> &str {
+        self.get(i)
+            .map(|m| m.as_str())
+            .unwrap_or_else(|| panic!("no group at index '{}'", i))
+    }
+}
+
+/// Copied from [`regex::Captures`]...
+///
+/// Get a group by name.
+///
+/// `'t` is the lifetime of the matched text and `'i` is the lifetime
+/// of the group name (the index).
+///
+/// The text can't outlive the `Captures` object if this method is
+/// used, because of how `Index` is defined (normally `a[i]` is part
+/// of `a` and can't outlive it); to do that, use `name` instead.
+///
+/// # Panics
+///
+/// If there is no group named by the given value.
+impl<'t, 'i> Index<&'i str> for Captures<'t> {
+    type Output = str;
+
+    fn index<'a>(&'a self, name: &'i str) -> &'a str {
+        self.name(name)
+            .map(|m| m.as_str())
+            .unwrap_or_else(|| panic!("no group named '{}'", name))
+    }
+}
+
 impl<'c, 't> Iterator for SubCaptureMatches<'c, 't> {
     type Item = Option<Match<'t>>;
 
@@ -631,6 +1007,8 @@ pub enum Expr {
     /// Back reference to a capture group, e.g. `\1` in `(abc|def)\1` references the captured group
     /// and the whole regex matches either `abcabc` or `defdef`.
     Backref(usize),
+    /// Back reference to a named capture group.
+    NamedBackref(String),
     /// Atomic non-capturing group, e.g. `(?>ab|a)` in text that contains `ab` will match `ab` and
     /// never backtrack and try `a`, even if matching fails after the atomic group.
     AtomicGroup(Box<Expr>),
@@ -649,6 +1027,26 @@ pub enum LookAround {
     LookBehindNeg,
 }
 
+/// An iterator over capture names in a [Regex].  The iterator
+/// returns the name of each group, or [None] if the group has
+/// no name.  Because capture group 0 cannot have a name, the
+/// first item returned is always [None].
+pub struct CaptureNames<'r>(std::vec::IntoIter<Option<&'r str>>);
+
+impl Debug for CaptureNames<'_> {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        f.write_str("<CaptureNames>")
+    }
+}
+
+impl<'r> Iterator for CaptureNames<'r> {
+    type Item = Option<&'r str>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        self.0.next()
+    }
+}
+
 // silly to write my own, but this is super-fast for the common 1-digit
 // case.
 fn push_usize(s: &mut String, x: usize) {
@@ -660,21 +1058,42 @@ fn push_usize(s: &mut String, x: usize) {
     }
 }
 
+fn is_special(c: char) -> bool {
+    match c {
+        '\\' | '.' | '+' | '*' | '?' | '(' | ')' | '|' | '[' | ']' | '{' | '}' | '^' | '$'
+        | '#' => true,
+        _ => false,
+    }
+}
+
 fn push_quoted(buf: &mut String, s: &str) {
     for c in s.chars() {
-        match c {
-            '\\' | '.' | '+' | '*' | '?' | '(' | ')' | '|' | '[' | ']' | '{' | '}' | '^' | '$'
-            | '#' => buf.push('\\'),
-            _ => (),
+        if is_special(c) {
+            buf.push('\\');
         }
         buf.push(c);
     }
 }
 
+/// Escapes special characters in `text` with '\\'.  Returns a string which, when interpreted
+/// as a regex, matches exactly `text`.
+pub fn escape(text: &str) -> Cow<str> {
+    // Using bytes() is OK because all special characters are single bytes.
+    match text.bytes().filter(|&b| is_special(b as char)).count() {
+        0 => Cow::Borrowed(text),
+        n => {
+            // The capacity calculation is exact because '\\' is a single byte.
+            let mut buf = String::with_capacity(text.len() + n);
+            push_quoted(&mut buf, text);
+            Cow::Owned(buf)
+        }
+    }
+}
+
 impl Expr {
     /// Parse the regex and return an expression (AST) and a bit set with the indexes of groups
     /// that are referenced by backrefs.
-    pub fn parse(re: &str) -> Result<(Expr, BitSet)> {
+    pub fn parse_tree(re: &str) -> Result<ExprTree> {
         Parser::parse(re)
     }
 
@@ -720,7 +1139,8 @@ impl Expr {
                     Expr::Empty => true,
                     _ => false,
                 };
-                let contains_empty = children.iter().any(&is_empty);
+
+                let contains_empty = children.iter().any(is_empty);
                 if contains_empty {
                     buf.push_str("(?:");
                 }
@@ -816,6 +1236,17 @@ fn codepoint_len(b: u8) -> usize {
     }
 }
 
+/// Returns the smallest possible index of the next valid UTF-8 sequence
+/// starting after `i`.
+/// Adapted from a function with the same name in the `regex` crate.
+fn next_utf8(text: &str, i: usize) -> usize {
+    let b = match text.as_bytes().get(i) {
+        None => return i + 1,
+        Some(&b) => b,
+    };
+    i + codepoint_len(b)
+}
+
 // If this returns false, then there is no possible backref in the re
 
 // Both potential implementations are turned off, because we currently
@@ -861,6 +1292,7 @@ mod tests {
     use crate::parse::make_literal;
     use crate::Expr;
     use crate::Regex;
+    use std::borrow::Cow;
     use std::usize;
     //use detect_possible_backref;
 
@@ -909,6 +1341,20 @@ mod tests {
         assert_eq!(s, format!("{:?}", regex));
     }
 
+    #[test]
+    fn display() {
+        let s = r"(a+)b\1";
+        let regex = Regex::new(s).unwrap();
+        assert_eq!(s, format!("{}", regex));
+    }
+
+    #[test]
+    fn from_str() {
+        let s = r"(a+)b\1";
+        let regex = s.parse::<Regex>().unwrap();
+        assert_eq!(regex.as_str(), s);
+    }
+
     #[test]
     fn to_str_repeat() {
         fn repeat(lo: usize, hi: usize, greedy: bool) -> Expr {
@@ -934,6 +1380,22 @@ mod tests {
         assert_eq!(to_str(repeat(1, usize::MAX, false)), "a+?");
     }
 
+    #[test]
+    fn escape() {
+        // Check that strings that need no quoting are borrowed, and that non-special punctuation
+        // is not quoted.
+        match crate::escape("@foo") {
+            Cow::Borrowed(s) => assert_eq!(s, "@foo"),
+            _ => panic!("Value should be borrowed."),
+        }
+
+        // Check typical usage.
+        assert_eq!(crate::escape("fo*o").into_owned(), "fo\\*o");
+
+        // Check that multibyte characters are handled correctly.
+        assert_eq!(crate::escape("fø*ø").into_owned(), "fø\\*ø");
+    }
+
     /*
     #[test]
     fn detect_backref() {
